/* Copyright (c) 2010-2011, Code Aurora Forum. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/mfd/pmic8058.h>
#include <linux/interrupt.h>
#include <linux/power_supply.h>
#include <linux/delay.h>
#include <linux/bitops.h>
#include <linux/debugfs.h>
#include <linux/msm-charger.h>
#include <linux/time.h>
#include <linux/slab.h>
#include <linux/wakelock.h>

#include <asm/atomic.h>

#include <mach/msm_hsusb.h>

// modified by yafi-begin
#if defined(ORG_VER)
#else
#include <mach/cci_hw_id.h>
#endif
// modified by yafi-end

#define MSM_CHG_MAX_EVENTS		16
#define CHARGING_TEOC_MS		9000000
#define UPDATE_TIME_MS			60000
#define RESUME_CHECK_PERIOD_MS		60000

#define DEFAULT_BATT_MAX_V		4200
#define DEFAULT_BATT_MIN_V		3200

#define MSM_CHARGER_GAUGE_MISSING_VOLTS 3500
#define MSM_CHARGER_GAUGE_MISSING_TEMP  35

// modified by yafi-begin
#if defined(ORG_VER)
#else
#define WAKE_LOCK_INITIALIZED            (1U << 8)

#define CONFIG_SMB137B_DEBUG 1
#if(CONFIG_SMB137B_DEBUG)
//    #define PrintLog_DEBUG(fmt, args...)    printk(KERN_INFO "CH(L)=> "pr_fmt(fmt), ##args)
    #define PrintLog_DEBUG(fmt, args...)    pr_debug("CH(L)=> "pr_fmt(fmt), ##args)
#else
    #define PrintLog_DEBUG(fmt, args...)
#endif

#define CONFIG_SMB137B_INFO 1
#if(CONFIG_SMB137B_INFO)
    #define PrintLog_INFO(fmt, args...)    printk(KERN_INFO "CH(L)=> "pr_fmt(fmt), ##args)
#else
    #define PrintLog_INFO(fmt, args...)
#endif
#endif
// modified by yafi-end

/**
 * enum msm_battery_status
 * @BATT_STATUS_ABSENT: battery not present
 * @BATT_STATUS_ID_INVALID: battery present but the id is invalid
 * @BATT_STATUS_DISCHARGING: battery is present and is discharging
 * @BATT_STATUS_TRKL_CHARGING: battery is being trickle charged
 * @BATT_STATUS_FAST_CHARGING: battery is being fast charged
 * @BATT_STATUS_JUST_FINISHED_CHARGING: just finished charging,
 *		battery is fully charged. Do not begin charging untill the
 *		voltage falls below a threshold to avoid overcharging
 * @BATT_STATUS_TEMPERATURE_OUT_OF_RANGE: battery present,
					no charging, temp is hot/cold
 */
enum msm_battery_status {
	BATT_STATUS_ABSENT,
	BATT_STATUS_ID_INVALID,
	BATT_STATUS_DISCHARGING,
	BATT_STATUS_TRKL_CHARGING,
	BATT_STATUS_FAST_CHARGING,
	BATT_STATUS_JUST_FINISHED_CHARGING,
	BATT_STATUS_TEMPERATURE_OUT_OF_RANGE,
};

struct msm_hardware_charger_priv {
	struct list_head list;
	struct msm_hardware_charger *hw_chg;
	enum msm_hardware_charger_state hw_chg_state;
	unsigned int max_source_current;
	struct power_supply psy;
// modified by yafi-begin
#if defined(ORG_VER)
#else
	struct power_supply psy_ac;
#endif
// modified by yafi-end
};

struct msm_charger_event {
	enum msm_hardware_charger_event event;
	struct msm_hardware_charger *hw_chg;
};

struct msm_charger_mux {
	int inited;
	struct list_head msm_hardware_chargers;
	int count_chargers;
	struct mutex msm_hardware_chargers_lock;

	struct device *dev;

	unsigned int max_voltage;
	unsigned int min_voltage;

	unsigned int safety_time;
	struct delayed_work teoc_work;

	unsigned int update_time;
	int stop_update;
	struct delayed_work update_heartbeat_work;
// modified by yafi-begin
#if defined(ORG_VER)
#else
	struct delayed_work update_battery_status_work;
#endif
// modified by yafi-end

	struct mutex status_lock;
	enum msm_battery_status batt_status;
	struct msm_hardware_charger_priv *current_chg_priv;
	struct msm_hardware_charger_priv *current_mon_priv;

	unsigned int (*get_batt_capacity_percent) (void);

	struct msm_charger_event *queue;
	int tail;
	int head;
	spinlock_t queue_lock;
	int queue_count;
	struct work_struct queue_work;
	struct workqueue_struct *event_wq_thread;
	struct wake_lock wl;
};

static struct msm_charger_mux msm_chg;

static struct msm_battery_gauge *msm_batt_gauge;

// modified by yafi-begin
#if defined(ORG_VER)
#else
static bool gmsm_charger_timeout_disable = false;
int last_percent = -1;
static bool over_reasonable_delta_percent = false;
int min_voltage_mv_DCHG = 4200;
extern bool is_charging_full;
extern bool is_charging_outof_temp;
extern bool is_charging_not_outof_temp;
extern bool is_charger_inserted;
#endif
// modified by yafi-end

static int is_chg_capable_of_charging(struct msm_hardware_charger_priv *priv)
{
	if (priv->hw_chg_state == CHG_READY_STATE
	    || priv->hw_chg_state == CHG_CHARGING_STATE)
		return 1;

	return 0;
}

static int is_batt_status_capable_of_charging(void)
{
	if (msm_chg.batt_status == BATT_STATUS_ABSENT
	    || msm_chg.batt_status == BATT_STATUS_TEMPERATURE_OUT_OF_RANGE
	    || msm_chg.batt_status == BATT_STATUS_ID_INVALID
	    || msm_chg.batt_status == BATT_STATUS_JUST_FINISHED_CHARGING)
		return 0;
	return 1;
}

static int is_batt_status_charging(void)
{
	if (msm_chg.batt_status == BATT_STATUS_TRKL_CHARGING
	    || msm_chg.batt_status == BATT_STATUS_FAST_CHARGING)
		return 1;
	return 0;
}

static int is_battery_present(void)
{
//	PrintLog_DEBUG("%s msm_batt_gauge02=0x%x\n", __func__, (int)msm_batt_gauge);

	if (msm_batt_gauge && msm_batt_gauge->is_battery_present)
		return msm_batt_gauge->is_battery_present();
	else {
		pr_err("msm-charger: no batt gauge batt=absent\n");
		return 0;
	}
}

static int is_battery_temp_within_range(void)
{
	if (msm_batt_gauge && msm_batt_gauge->is_battery_temp_within_range)
		return msm_batt_gauge->is_battery_temp_within_range();
	else {
		pr_err("msm-charger no batt gauge batt=out_of_temperatur\n");
		return 0;
	}
}

static int is_battery_id_valid(void)
{
	if (msm_batt_gauge && msm_batt_gauge->is_battery_id_valid)
		return msm_batt_gauge->is_battery_id_valid();
	else {
		pr_err("msm-charger no batt gauge batt=id_invalid\n");
		return 0;
	}
}

static int get_prop_battery_mvolts(void)
{
	if (msm_batt_gauge && msm_batt_gauge->get_battery_mvolts)
		return msm_batt_gauge->get_battery_mvolts();
	else {
		pr_err("msm-charger no batt gauge assuming 3.5V\n");
		return MSM_CHARGER_GAUGE_MISSING_VOLTS;
	}
}

static int get_battery_temperature(void)
{
	if (msm_batt_gauge && msm_batt_gauge->get_battery_temperature)
		return msm_batt_gauge->get_battery_temperature();
	else {
		pr_err("msm-charger no batt gauge assuming 35 deg G\n");
		return MSM_CHARGER_GAUGE_MISSING_TEMP;
	}
}

static int get_prop_batt_capacity(void)
{
	if (msm_batt_gauge && msm_batt_gauge->get_batt_remaining_capacity)
		return msm_batt_gauge->get_batt_remaining_capacity();

	return msm_chg.get_batt_capacity_percent();
}

static int get_prop_batt_health(void)
{
	int status = 0;

	if (msm_chg.batt_status == BATT_STATUS_TEMPERATURE_OUT_OF_RANGE)
		status = POWER_SUPPLY_HEALTH_OVERHEAT;
	else
		status = POWER_SUPPLY_HEALTH_GOOD;

	return status;
}

static int get_prop_charge_type(void)
{
	int status = 0;

	if (msm_chg.batt_status == BATT_STATUS_TRKL_CHARGING)
		status = POWER_SUPPLY_CHARGE_TYPE_TRICKLE;
	else if (msm_chg.batt_status == BATT_STATUS_FAST_CHARGING)
		status = POWER_SUPPLY_CHARGE_TYPE_FAST;
	else
		status = POWER_SUPPLY_CHARGE_TYPE_NONE;

	return status;
}

static int get_prop_batt_status(void)
{
	int status = 0;

	if (msm_batt_gauge && msm_batt_gauge->get_battery_status) {
		status = msm_batt_gauge->get_battery_status();
		if (status == POWER_SUPPLY_STATUS_CHARGING ||
			status == POWER_SUPPLY_STATUS_FULL ||
			status == POWER_SUPPLY_STATUS_DISCHARGING)
			return status;
	}

	if (is_batt_status_charging())
		status = POWER_SUPPLY_STATUS_CHARGING;
	else if (msm_chg.batt_status ==
		 BATT_STATUS_JUST_FINISHED_CHARGING
			 && msm_chg.current_chg_priv != NULL)
		status = POWER_SUPPLY_STATUS_FULL;
	else
		status = POWER_SUPPLY_STATUS_DISCHARGING;

	return status;
}

 /* This function should only be called within handle_event or resume */
static void update_batt_status(void)
{
	PrintLog_DEBUG("%s batt_status=%d\n", __func__, msm_chg.batt_status);
	
	if (is_battery_present()) {
		if (is_battery_id_valid()) {
			if (msm_chg.batt_status == BATT_STATUS_ABSENT
				|| msm_chg.batt_status
					== BATT_STATUS_ID_INVALID) {
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
			}
		} else
			msm_chg.batt_status = BATT_STATUS_ID_INVALID;
	 } else
		msm_chg.batt_status = BATT_STATUS_ABSENT;
}

static enum power_supply_property msm_power_props[] = {
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_ONLINE,
};

static char *msm_power_supplied_to[] = {
	"battery",
};

static int msm_power_get_property(struct power_supply *psy,
				  enum power_supply_property psp,
				  union power_supply_propval *val)
{
	struct msm_hardware_charger_priv *priv;
// modified by yafi-begin
#if defined(ORG_VER)
#else
	char *pstr_prop = NULL;
	char *pstr_content = NULL;
#endif
// modified by yafi-end

	priv = container_of(psy, struct msm_hardware_charger_priv, psy);
	switch (psp) {
	case POWER_SUPPLY_PROP_PRESENT:
// modified by yafi-begin
#if defined(ORG_VER)
		val->intval = !(priv->hw_chg_state == CHG_ABSENT_STATE);
#else
		if (priv->max_source_current != 1800)
			val->intval = !(priv->hw_chg_state == CHG_ABSENT_STATE);
		else
			val->intval = 0;
		pstr_prop = "USB_PRESENT";
		if(val->intval == 1)
			pstr_content = "PRESENT";
		else
			pstr_content = "ABSENT";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_ONLINE:
// modified by yafi-begin
#if defined(ORG_VER)
		val->intval = (priv->hw_chg_state == CHG_READY_STATE)
			|| (priv->hw_chg_state == CHG_CHARGING_STATE);
#else
		if (priv->max_source_current != 1800)
			val->intval = !(priv->hw_chg_state == CHG_ABSENT_STATE);
		else
			val->intval = 0;
		pstr_prop = "USB_ONLINE";
		if(val->intval == 1)
			pstr_content = "PRESENT";
		else
			pstr_content = "ABSENT";
#endif
// modified by yafi-end
		break;
	default:
		return -EINVAL;
	}

// modified by yafi-begin
#if defined(ORG_VER)
#else
	switch(psp)
	{
	case POWER_SUPPLY_PROP_ONLINE:
//		PrintLog_INFO("%s=%s\n", pstr_prop, pstr_content);
		PrintLog_DEBUG("%s=%s\n", pstr_prop, pstr_content);
		break;
	default:
		break;
	}
#endif
// modified by yafi-end
	return 0;
}

// modified by yafi-begin
#if defined(ORG_VER)
#else
static int msm_power_get_property_ac(struct power_supply *psy_ac,
				  enum power_supply_property psp,
				  union power_supply_propval *val) //YaFi-OK
{
	struct msm_hardware_charger_priv *priv;
	char *pstr_prop = NULL;
	char *pstr_content = NULL;

	priv = container_of(psy_ac, struct msm_hardware_charger_priv, psy_ac);
	switch (psp) {
	case POWER_SUPPLY_PROP_PRESENT:
		if (priv->max_source_current == 1800)
			val->intval = !(priv->hw_chg_state == CHG_ABSENT_STATE);
		else
			val->intval = 0;
		pstr_prop = "AC_PRESENT";
		if(val->intval == 1)
			pstr_content = "PRESENT";
		else
			pstr_content = "ABSENT";
		break;
	case POWER_SUPPLY_PROP_ONLINE:
		if (priv->max_source_current == 1800)
			val->intval = !(priv->hw_chg_state == CHG_ABSENT_STATE);
		else
			val->intval = 0;
		pstr_prop = "AC_ONLINE";
		if(val->intval == 1)
			pstr_content = "PRESENT";
		else
			pstr_content = "ABSENT";
		break;
	default:
		return -EINVAL;
	}

	switch(psp)
	{
	case POWER_SUPPLY_PROP_ONLINE:
//		PrintLog_INFO("%s=%s\n", pstr_prop, pstr_content);
		PrintLog_DEBUG("%s=%s\n", pstr_prop, pstr_content);
		break;
	default:
		break;
	}
	return 0;
}
#endif
// modified by yafi-end

static enum power_supply_property msm_batt_power_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_CHARGE_TYPE,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_TECHNOLOGY,
	POWER_SUPPLY_PROP_VOLTAGE_MAX_DESIGN,
	POWER_SUPPLY_PROP_VOLTAGE_MIN_DESIGN,
	POWER_SUPPLY_PROP_VOLTAGE_NOW,
	POWER_SUPPLY_PROP_CAPACITY,
// modified by yafi-begin
#if defined(ORG_VER)
#else
	POWER_SUPPLY_PROP_TEMP,
#endif
// modified by yafi-end
};

static int msm_batt_power_get_property(struct power_supply *psy,
				       enum power_supply_property psp,
				       union power_supply_propval *val)
{
// modified by yafi-begin
#if defined(ORG_VER)
#else
	char *pstr_prop = NULL;
	char *pstr_content = NULL;
	char string[5];
#endif
// modified by yafi-end

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
		val->intval = get_prop_batt_status();
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_STATUS";
		if(val->intval == POWER_SUPPLY_STATUS_CHARGING)
			pstr_content = "CHARGING";
		else if(val->intval == POWER_SUPPLY_STATUS_FULL)
			pstr_content = "FULL";
		else
			pstr_content = "DISCHARGING";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_CHARGE_TYPE:
		val->intval = get_prop_charge_type();
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_TYPE";
		if(val->intval == POWER_SUPPLY_CHARGE_TYPE_TRICKLE)
			pstr_content = "TRICKLE";
		else if(val->intval == POWER_SUPPLY_CHARGE_TYPE_FAST)
			pstr_content = "FAST";
		else
			pstr_content = "NONE";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_HEALTH:
		val->intval = get_prop_batt_health();
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_HEALTH";
		if(val->intval == POWER_SUPPLY_HEALTH_OVERHEAT)
			pstr_content = "OVERHEAT";
		else
			pstr_content = "GOOD";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_PRESENT:
		val->intval = !(msm_chg.batt_status == BATT_STATUS_ABSENT);
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_PRESENT";
		if(val->intval == 1)
			pstr_content = "PRESENT";
		else
			pstr_content = "ABSENT";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_TECHNOLOGY:
		val->intval = POWER_SUPPLY_TECHNOLOGY_NiMH;
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_TECHNOLOGY";
		pstr_content = "NiMH";
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_MAX_DESIGN:
		val->intval = msm_chg.max_voltage * 1000;
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_VOLTAGE_MAX_DESIGN";
		sprintf(string, "%d", val->intval/1000);
		pstr_content = string;
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_MIN_DESIGN:
		val->intval = msm_chg.min_voltage * 1000;
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_VOLTAGE_MIN_DESIGN";
		sprintf(string, "%d", val->intval/1000);
		pstr_content = string;
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
// modified by yafi-begin
#if defined(ORG_VER)
		val->intval = get_prop_battery_mvolts();
#else
		val->intval = get_prop_battery_mvolts() * 1000;
		pstr_prop = "BAT_VOLTAGE";
		sprintf(string, "%d", val->intval/1000);
		pstr_content = string;
#endif
// modified by yafi-end
		break;
	case POWER_SUPPLY_PROP_CAPACITY:
		val->intval = get_prop_batt_capacity();
// modified by yafi-begin
#if defined(ORG_VER)
#else
		pstr_prop = "BAT_CAPACITY";
		sprintf(string, "%d", val->intval);
		pstr_content = string;
#endif
// modified by yafi-end
		break;
// modified by yafi-begin
#if defined(ORG_VER)
#else
	case POWER_SUPPLY_PROP_TEMP:
		val->intval = get_battery_temperature();
		pstr_prop = "BAT_TEMP";
		sprintf(string, "%d", val->intval);
		pstr_content = string;
		break;
#endif
// modified by yafi-end
	default:
		return -EINVAL;
	}

// modified by yafi-begin
#if defined(ORG_VER)
#else
	switch(psp)
	{
	case POWER_SUPPLY_PROP_STATUS:
	case POWER_SUPPLY_PROP_CHARGE_TYPE:
	case POWER_SUPPLY_PROP_HEALTH:
	case POWER_SUPPLY_PROP_VOLTAGE_NOW:
	case POWER_SUPPLY_PROP_CAPACITY:
	case POWER_SUPPLY_PROP_TEMP:
//		PrintLog_INFO("%s=%s\n", pstr_prop, pstr_content);
		PrintLog_DEBUG("%s=%s\n", pstr_prop, pstr_content);
		break;
	default:
		break;
	}
#endif
// modified by yafi-end
	return 0;
}

static struct power_supply msm_psy_batt = {
	.name = "battery",
	.type = POWER_SUPPLY_TYPE_BATTERY,
	.properties = msm_batt_power_props,
	.num_properties = ARRAY_SIZE(msm_batt_power_props),
	.get_property = msm_batt_power_get_property,
};

static int usb_chg_current;
static struct msm_hardware_charger_priv *usb_hw_chg_priv;
static void (*notify_vbus_state_func_ptr)(int);
static int usb_notified_of_insertion;

/* this is passed to the hsusb via platform_data msm_otg_pdata */
int msm_charger_register_vbus_sn(void (*callback)(int))
{
	pr_debug(KERN_INFO "%s\n", __func__);
	PrintLog_DEBUG("%s\n", __func__);
	notify_vbus_state_func_ptr = callback;
	return 0;
}

/* this is passed to the hsusb via platform_data msm_otg_pdata */
void msm_charger_unregister_vbus_sn(void (*callback)(int))
{
	pr_debug(KERN_INFO "%s\n", __func__);
	PrintLog_DEBUG("%s\n", __func__);
	notify_vbus_state_func_ptr = NULL;
}

// Bug 909, Luke Implement MHL charging function
char cci_usb_power_plugin = 0;
static void notify_usb_of_the_plugin_event(struct msm_hardware_charger_priv
					   *hw_chg, int plugin)
{
	plugin = !!plugin;
	if (plugin == 1 && usb_notified_of_insertion == 0) {
		usb_notified_of_insertion = 1;
// modified by yafi-begin
#if defined(ORG_VER)
#else
		usb_hw_chg_priv = hw_chg;
#endif
// modified by yafi-end
		if (notify_vbus_state_func_ptr) {
// modified by yafi-begin
#if defined(ORG_VER)
			dev_err(msm_chg.dev, "%s notifying plugin\n", __func__);
#else
			PrintLog_INFO("%s notifying plugin\n", __func__);
#endif
// modified by yafi-end
			(*notify_vbus_state_func_ptr) (plugin);
		} else
			dev_err(msm_chg.dev, "%s unable to notify plugin\n",
				__func__);
// modified by yafi-begin
#if defined(ORG_VER)
		usb_hw_chg_priv = hw_chg;
#else
#endif
// modified by yafi-end
                // Bug 909, Luke Implement MHL charging function
		cci_usb_power_plugin = 1;

	}
	if (plugin == 0 && usb_notified_of_insertion == 1) {
		if (notify_vbus_state_func_ptr) {
// modified by yafi-begin
#if defined(ORG_VER)
			dev_err(msm_chg.dev, "%s notifying unplugin\n",
				__func__);
#else
			PrintLog_INFO("%s notifying unplugin\n", __func__);
#endif
// modified by yafi-end
			(*notify_vbus_state_func_ptr) (plugin);
		} else
			dev_err(msm_chg.dev, "%s unable to notify unplugin\n",
				__func__);
// modified by yafi-begin
#if defined(ORG_VER)
		usb_notified_of_insertion = 0;
		usb_hw_chg_priv = NULL;
#else
		usb_hw_chg_priv = NULL;
		usb_notified_of_insertion = 0;
#endif
// modified by yafi-end

                // Bug 909, Luke Implement MHL charging function
                cci_usb_power_plugin = 0;

	}
}

// modified by yafi-begin
#if defined(ORG_VER)
#else
struct Vbat_mv_VS_Vbat_capacity
{
	unsigned int Vbat_mv;
	unsigned int Vbat_capacity;
};

#define VBAT_ADC_MV_OFFSET 275
static struct Vbat_mv_VS_Vbat_capacity Vbat_mv_capacity[]=
{
{3200,1},{3267,2},{3333,3},{3400,4},{3467,5},
{3533,6},{3600,7},{3606,8},{3612,9},{3618,10},
{3624,11},{3630,12},{3636,13},{3642,14},{3648,15},
{3654,16},{3659,17},{3665,18},{3671,19},{3677,20},
{3683,21},{3689,22},{3695,23},{3701,24},{3707,25},
{3713,26},{3719,27},{3725,28},{3731,29},{3737,30},
{3743,31},{3749,32},{3755,33},{3761,34},{3766,35},
{3772,36},{3778,37},{3784,38},{3790,39},{3796,40},
{3802,41},{3808,42},{3814,43},{3820,44},{3826,45},
{3832,46},{3838,47},{3844,48},{3850,49},{3856,50},
{3862,51},{3867,52},{3873,53},{3879,54},{3885,55},
{3891,56},{3897,57},{3903,58},{3909,59},{3915,60},
{3921,61},{3927,62},{3933,63},{3939,64},{3945,65},
{3951,66},{3957,67},{3963,68},{3969,69},{3974,70},
{3980,71},{3986,72},{3992,73},{3998,74},{4004,75},
{4010,76},{4016,77},{4022,78},{4028,79},{4034,80},
{4040,81},{4046,82},{4052,83},{4058,84},{4064,85},
{4070,86},{4076,87},{4081,88},{4087,89},{4093,90},
{4099,91},{4105,92},{4111,93},{4117,94},{4123,95},
{4129,96},{4135,97},{4150,98},{4165,99},{4200,100},
};

static unsigned int get_Vbat_capacity(struct Vbat_mv_VS_Vbat_capacity *pVbat_capacity_table, unsigned int entry_count, unsigned int Vbat_mv)
{
	int Vbat_mv_capacity_count = entry_count;
	int index_low = 0, index_high = 0, i = 0;

	if (Vbat_mv<pVbat_capacity_table[0].Vbat_mv)
		return 0;
	else if(Vbat_mv>pVbat_capacity_table[Vbat_mv_capacity_count-1].Vbat_mv)
		return 100;
	
	for (i=0; i<Vbat_mv_capacity_count; i++)
	{
		if (Vbat_mv >= pVbat_capacity_table[i].Vbat_mv)
		{
			if (Vbat_mv == pVbat_capacity_table[i].Vbat_mv)
				return pVbat_capacity_table[i].Vbat_capacity;
			index_low = i;
			continue;
		}
		else
		{
			index_high = i;
			break;
		}
	}

	return pVbat_capacity_table[index_low].Vbat_capacity;
}
#endif
// modified by yafi-end

extern bool smb137b_is_charging_full(void);
#define AVG_SAMPLE_VBAT_PERCENT_NUM 1
#define DEFAULT_PERCENT 50
#define REASONABLE_DELTA_PERCENT (3)
static unsigned int msm_chg_get_batt_capacity_percent(void)
{
// modified by yafi-begin
#if defined(ORG_VER)
	unsigned int current_voltage = get_prop_battery_mvolts();
	unsigned int low_voltage = msm_chg.min_voltage;
	unsigned int high_voltage = msm_chg.max_voltage;

	if (current_voltage <= low_voltage)
		return 0;
	else if (current_voltage >= high_voltage)
		return 100;
	else
		return (current_voltage - low_voltage) * 100
		    / (high_voltage - low_voltage);
#else
	static int avg_sample_vbatt_percent[AVG_SAMPLE_VBAT_PERCENT_NUM];
	static int avg_sample_vbatt_percent_index = 0;
	static bool avg_sample_vbatt_percent_recycle = false;
	static unsigned int full_voltage = 0;
	int i, sum_vbatt_percent = 0;
	unsigned int current_voltage = 0;
	unsigned int current_percent = 0;
	
	current_voltage = get_prop_battery_mvolts();
	if (current_voltage == 0)
	{
		if (last_percent < 0)
			current_percent = DEFAULT_PERCENT;
		else
			current_percent = last_percent;
	}
	else
	{
		if ((msm_chg.batt_status == BATT_STATUS_JUST_FINISHED_CHARGING && msm_chg.current_chg_priv != NULL) || 
			(full_voltage != 0 && current_voltage >= full_voltage))
		{
			current_percent = 100;
			for(i=0;i<AVG_SAMPLE_VBAT_PERCENT_NUM;i++)
				avg_sample_vbatt_percent[i] = current_percent;
			avg_sample_vbatt_percent_recycle = true;
			if(full_voltage)
				full_voltage = (full_voltage + current_voltage)/2;
			else
				full_voltage = current_voltage;
		}
		else
		{
			if (is_charger_inserted)
			{
			}
			else
			{
				if (current_voltage > min_voltage_mv_DCHG)
					current_voltage = min_voltage_mv_DCHG;
				else
					min_voltage_mv_DCHG = current_voltage;
			}
			current_percent = get_Vbat_capacity((struct Vbat_mv_VS_Vbat_capacity *)(&Vbat_mv_capacity), (unsigned int)(sizeof(Vbat_mv_capacity)/sizeof(Vbat_mv_capacity[0])), current_voltage);
		}

		avg_sample_vbatt_percent[avg_sample_vbatt_percent_index] = current_percent;
		if (++avg_sample_vbatt_percent_index >= AVG_SAMPLE_VBAT_PERCENT_NUM)
		{
			avg_sample_vbatt_percent_index = 0;
			avg_sample_vbatt_percent_recycle = true;
		}
		if(avg_sample_vbatt_percent_recycle)
		{	
			for(i=0;i<AVG_SAMPLE_VBAT_PERCENT_NUM;i++)
				sum_vbatt_percent += avg_sample_vbatt_percent[i];
			current_percent = sum_vbatt_percent/AVG_SAMPLE_VBAT_PERCENT_NUM;
		}
		else
		{
			for(i=0;i<avg_sample_vbatt_percent_index;i++)
				sum_vbatt_percent += avg_sample_vbatt_percent[i];
			current_percent = sum_vbatt_percent/avg_sample_vbatt_percent_index;
		}
// modified by yafi-begin
#if defined(ORG_VER)
		last_percent = current_percent;
#else
		if (last_percent < 0)
			last_percent = current_percent;
		else
		{
			if(over_reasonable_delta_percent)
			{
				current_percent = last_percent;
			}
			else
			{
				if (current_percent > (last_percent + REASONABLE_DELTA_PERCENT))
				{
					last_percent = current_percent = (last_percent + REASONABLE_DELTA_PERCENT);
					over_reasonable_delta_percent = true;
				}
				else if ((int)current_percent < (int)(last_percent - REASONABLE_DELTA_PERCENT))
				{
					last_percent = current_percent = (last_percent - REASONABLE_DELTA_PERCENT)>0?(last_percent - REASONABLE_DELTA_PERCENT):0;
					over_reasonable_delta_percent = true;
				}
				else
				{
					last_percent = current_percent;
				}
			}
		}
#endif
// modified by yafi-end
	}

	PrintLog_DEBUG("%s Vbat_mv=%d, Vbat_percent=%d\n", __func__, current_voltage, current_percent);

	return current_percent;
#endif
// modified by yafi-end
}

#ifdef DEBUG
static inline void debug_print(const char *func,
			       struct msm_hardware_charger_priv *hw_chg_priv)
{
	dev_dbg(msm_chg.dev,
		"%s current=(%s)(s=%d)(r=%d) new=(%s)(s=%d)(r=%d) batt=%d En\n",
		func,
		msm_chg.current_chg_priv ? msm_chg.current_chg_priv->
		hw_chg->name : "none",
		msm_chg.current_chg_priv ? msm_chg.
		current_chg_priv->hw_chg_state : -1,
		msm_chg.current_chg_priv ? msm_chg.current_chg_priv->
		hw_chg->rating : -1,
		hw_chg_priv ? hw_chg_priv->hw_chg->name : "none",
		hw_chg_priv ? hw_chg_priv->hw_chg_state : -1,
		hw_chg_priv ? hw_chg_priv->hw_chg->rating : -1,
		msm_chg.batt_status);
}
#else
static inline void debug_print(const char *func,
			       struct msm_hardware_charger_priv *hw_chg_priv)
{
	PrintLog_DEBUG("%s current=(%s)(s=%d)(r=%d) new=(%s)(s=%d)(r=%d) batt=%d En\n",
		func,
		msm_chg.current_chg_priv ? msm_chg.current_chg_priv->
		hw_chg->name : "none",
		msm_chg.current_chg_priv ? msm_chg.
		current_chg_priv->hw_chg_state : -1,
		msm_chg.current_chg_priv ? msm_chg.current_chg_priv->
		hw_chg->rating : -1,
		hw_chg_priv ? hw_chg_priv->hw_chg->name : "none",
		hw_chg_priv ? hw_chg_priv->hw_chg_state : -1,
		hw_chg_priv ? hw_chg_priv->hw_chg->rating : -1,
		msm_chg.batt_status);
}
#endif

static struct msm_hardware_charger_priv *find_best_charger(void)
{
	struct msm_hardware_charger_priv *hw_chg_priv;
	struct msm_hardware_charger_priv *better;
	int rating;

	better = NULL;
	rating = 0;

	list_for_each_entry(hw_chg_priv, &msm_chg.msm_hardware_chargers, list) {
		if (is_chg_capable_of_charging(hw_chg_priv)) {
			if (hw_chg_priv->hw_chg->rating > rating) {
				rating = hw_chg_priv->hw_chg->rating;
				better = hw_chg_priv;
			}
		}
	}

	return better;
}

static int msm_charging_switched(struct msm_hardware_charger_priv *priv)
{
	int ret = 0;

	if (priv->hw_chg->charging_switched)
		ret = priv->hw_chg->charging_switched(priv->hw_chg);
	return ret;
}

static int msm_stop_charging(struct msm_hardware_charger_priv *priv)
{
	int ret;

	ret = priv->hw_chg->stop_charging(priv->hw_chg);
// modified by yafi-begin
#if defined(ORG_VER)
	if (!ret)
#else
#endif
// modified by yafi-end
		wake_unlock(&msm_chg.wl);
	return ret;
}

// modified by yafi-begin
#if defined(ORG_VER)
static void msm_enable_system_current(struct msm_hardware_charger_priv *priv)
{
	if (priv->hw_chg->start_system_current)
		priv->hw_chg->start_system_current(priv->hw_chg,
					 priv->max_source_current);
}

static void msm_disable_system_current(struct msm_hardware_charger_priv *priv)
{
	if (priv->hw_chg->stop_system_current)
		priv->hw_chg->stop_system_current(priv->hw_chg);
}
#else
#endif
// modified by yafi-end

/* the best charger has been selected -start charging from current_chg_priv */
static int msm_start_charging(void)
{
	int ret;
	struct msm_hardware_charger_priv *priv;

	priv = msm_chg.current_chg_priv;

	PrintLog_DEBUG("%s, %s chg_voltage=%d chg_current=%d",
		__func__,priv->hw_chg->name, msm_chg.max_voltage, priv->max_source_current);
	
	wake_lock(&msm_chg.wl);
	ret = priv->hw_chg->start_charging(priv->hw_chg, msm_chg.max_voltage,
					 priv->max_source_current);
	if (ret) {
		wake_unlock(&msm_chg.wl);
		dev_err(msm_chg.dev, "%s couldnt start chg error = %d\n",
			priv->hw_chg->name, ret);
	} else
		priv->hw_chg_state = CHG_CHARGING_STATE;

	return ret;
}

static void handle_charging_done(struct msm_hardware_charger_priv *priv)
{
	if (msm_chg.current_chg_priv == priv) {
		if (msm_chg.current_chg_priv->hw_chg_state ==
		    CHG_CHARGING_STATE)
			if (msm_stop_charging(msm_chg.current_chg_priv)) {
				dev_err(msm_chg.dev, "%s couldnt stop chg\n",
					msm_chg.current_chg_priv->hw_chg->name);
			}
		msm_chg.current_chg_priv->hw_chg_state = CHG_READY_STATE;

		msm_chg.batt_status = BATT_STATUS_JUST_FINISHED_CHARGING;
		dev_info(msm_chg.dev, "%s: stopping safety timer work\n",
				__func__);
		cancel_delayed_work(&msm_chg.teoc_work);
// modified by yafi-begin
#if defined(ORG_VER)
		if (msm_batt_gauge && msm_batt_gauge->monitor_for_recharging)
			msm_batt_gauge->monitor_for_recharging();
		else
			dev_err(msm_chg.dev,
			      "%s: no batt gauge recharge monitor\n", __func__);
#else
#endif
// modified by yafi-end
	}
}

static void teoc(struct work_struct *work)
{
	/* we have been charging too long - stop charging */
	dev_info(msm_chg.dev, "%s: safety timer work expired\n", __func__);

	mutex_lock(&msm_chg.status_lock);
	if (msm_chg.current_chg_priv != NULL
	    && msm_chg.current_chg_priv->hw_chg_state == CHG_CHARGING_STATE) {
		handle_charging_done(msm_chg.current_chg_priv);
	}
	mutex_unlock(&msm_chg.status_lock);
}

static void handle_battery_inserted(void)
{
	PrintLog_DEBUG("%s current_chg_priv=0x%x\n", __func__, (int)msm_chg.current_chg_priv);
	/* if a charger is already present start charging */
	if (msm_chg.current_chg_priv != NULL &&
	    is_batt_status_capable_of_charging() &&
	    !is_batt_status_charging()) {
		if (msm_start_charging()) {
			dev_err(msm_chg.dev, "%s couldnt start chg\n",
				msm_chg.current_chg_priv->hw_chg->name);
			return;
		}
		msm_chg.batt_status = BATT_STATUS_TRKL_CHARGING;

		dev_info(msm_chg.dev, "%s: starting safety timer work\n",
				__func__);
// modified by yafi-begin
#if defined(ORG_VER)
#else
		if (!gmsm_charger_timeout_disable)
#endif
// modified by yafi-end
		queue_delayed_work(msm_chg.event_wq_thread,
					&msm_chg.teoc_work,
				      round_jiffies_relative(msecs_to_jiffies
							     (msm_chg.
							      safety_time)));
	}
}

static void handle_battery_removed(void)
{
	/* if a charger is charging the battery stop it */
	if (msm_chg.current_chg_priv != NULL
	    && msm_chg.current_chg_priv->hw_chg_state == CHG_CHARGING_STATE) {
		if (msm_stop_charging(msm_chg.current_chg_priv)) {
			dev_err(msm_chg.dev, "%s couldnt stop chg\n",
				msm_chg.current_chg_priv->hw_chg->name);
		}
		msm_chg.current_chg_priv->hw_chg_state = CHG_READY_STATE;

		dev_info(msm_chg.dev, "%s: stopping safety timer work\n",
				__func__);
		cancel_delayed_work(&msm_chg.teoc_work);
	}
}

// modified by yafi-begin
#if defined(ORG_VER)
#else
extern int smb137b_get_battery_mvolts(void);

void update_power_supply_batt(void)
{
	smb137b_get_battery_mvolts();
	/* notify that the voltage has changed
	 * the read of the capacity will trigger a
	 * voltage read*/
	power_supply_changed(&msm_psy_batt);
}

static void update_battery_status(struct work_struct *work)
{
	/* notify that the voltage has changed
	 * the read of the capacity will trigger a
	 * voltage read*/
	power_supply_changed(&msm_psy_batt);
}
#endif
// modified by yafi-end

static void update_heartbeat(struct work_struct *work)
{
	int temperature;

	PrintLog_DEBUG("%s batt_status-01= %d\n", __func__, msm_chg.batt_status);

	if (msm_chg.batt_status == BATT_STATUS_ABSENT
		|| msm_chg.batt_status == BATT_STATUS_ID_INVALID) {
		if (is_battery_present())
			if (is_battery_id_valid()) {
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
				handle_battery_inserted();
			}
	} else {
		if (!is_battery_present()) {
			msm_chg.batt_status = BATT_STATUS_ABSENT;
			handle_battery_removed();
		}
		/*
		 * check battery id because a good battery could be removed
		 * and replaced with a invalid battery.
		 */
		if (!is_battery_id_valid()) {
			msm_chg.batt_status = BATT_STATUS_ID_INVALID;
			handle_battery_removed();
		}
	}
	pr_debug("msm-charger %s batt_status= %d\n",
				__func__, msm_chg.batt_status);
	PrintLog_DEBUG("%s batt_status-02= %d\n",
				__func__, msm_chg.batt_status);

	if (msm_chg.current_chg_priv
		&& msm_chg.current_chg_priv->hw_chg_state
			== CHG_CHARGING_STATE) {
		temperature = get_battery_temperature();
		/* TODO implement JEITA SPEC*/
	}

// modified by yafi-begin
#if defined(ORG_VER)
	/* notify that the voltage has changed
	 * the read of the capacity will trigger a
	 * voltage read*/
	power_supply_changed(&msm_psy_batt);
#else
	smb137b_get_battery_mvolts();
	over_reasonable_delta_percent = false;
    power_supply_changed(&msm_psy_batt);
#endif
// modified by yafi-end

	if (msm_chg.stop_update) {
		msm_chg.stop_update = 0;
		return;
	}
	queue_delayed_work(msm_chg.event_wq_thread,
				&msm_chg.update_heartbeat_work,
			      round_jiffies_relative(msecs_to_jiffies
						     (msm_chg.update_time)));
}

/* set the charger state to READY before calling this */
static void handle_charger_ready(struct msm_hardware_charger_priv *hw_chg_priv)
{
// modified by yafi-begin
#if defined(ORG_VER)
	struct msm_hardware_charger_priv *old_chg_priv = NULL;
#else
#endif
// modified by yafi-end

//	debug_print(__func__, hw_chg_priv);

	PrintLog_DEBUG("%s current_chg_priv=0x%x", __func__, (int)msm_chg.current_chg_priv);
	if((int)msm_chg.current_chg_priv)
		PrintLog_DEBUG("%s msm_chg:0x%x,%d,%d hw_chg_priv:%d\n", __func__, \
		(int)msm_chg.current_chg_priv, msm_chg.current_chg_priv->hw_chg->rating, msm_chg.current_chg_priv->hw_chg_state, \
		hw_chg_priv->hw_chg->rating);

// modified by yafi-begin
#if defined(ORG_VER)
	if (msm_chg.current_chg_priv != NULL
	    && hw_chg_priv->hw_chg->rating >
	    msm_chg.current_chg_priv->hw_chg->rating) {
#else
	if (msm_chg.current_chg_priv != NULL
	    && hw_chg_priv->hw_chg->rating >=
	    msm_chg.current_chg_priv->hw_chg->rating) {
#endif
// modified by yafi-end
		/*
		 * a better charger was found, ask the current charger
		 * to stop charging if it was charging
		 */
		if (msm_chg.current_chg_priv->hw_chg_state ==
		    CHG_CHARGING_STATE) {
			if (msm_stop_charging(msm_chg.current_chg_priv)) {
				dev_err(msm_chg.dev, "%s couldnt stop chg\n",
					msm_chg.current_chg_priv->hw_chg->name);
				return;
			}
			if (msm_charging_switched(msm_chg.current_chg_priv)) {
				dev_err(msm_chg.dev, "%s couldnt switch chg\n",
					msm_chg.current_chg_priv->hw_chg->name);
				return;
			}
		}
		msm_chg.current_chg_priv->hw_chg_state = CHG_READY_STATE;
// modified by yafi-begin
#if defined(ORG_VER)
		old_chg_priv = msm_chg.current_chg_priv;
#else
#endif
// modified by yafi-end
		msm_chg.current_chg_priv = NULL;
	}

	if (msm_chg.current_chg_priv == NULL) {
		msm_chg.current_chg_priv = hw_chg_priv;
		dev_info(msm_chg.dev,
			 "%s: best charger = %s\n", __func__,
			 msm_chg.current_chg_priv->hw_chg->name);

// modified by yafi-begin
#if defined(ORG_VER)
		msm_enable_system_current(msm_chg.current_chg_priv);
		/*
		 * since a better charger was chosen, ask the old
		 * charger to stop providing system current
		 */
		if (old_chg_priv != NULL)
			msm_disable_system_current(old_chg_priv);
#else
#endif
// modified by yafi-end

		if (!is_batt_status_capable_of_charging())
			return;

		/* start charging from the new charger */
		if (!msm_start_charging()) {
			/* if we simply switched chg continue with teoc timer
			 * else we update the batt state and set the teoc
			 * timer */
			if (!is_batt_status_charging()) {
				dev_info(msm_chg.dev,
				       "%s: starting safety timer\n", __func__);
// modified by yafi-begin
#if defined(ORG_VER)
#else
				if (!gmsm_charger_timeout_disable)
#endif
// modified by yafi-end
				queue_delayed_work(msm_chg.event_wq_thread,
							&msm_chg.teoc_work,
						      round_jiffies_relative
						      (msecs_to_jiffies
						       (msm_chg.safety_time)));
				msm_chg.batt_status = BATT_STATUS_TRKL_CHARGING;
			}
		} else {
			/* we couldnt start charging from the new readied
			 * charger */
			if (is_batt_status_charging())
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
		}
	}
}

static void handle_charger_removed(struct msm_hardware_charger_priv
				   *hw_chg_removed, int new_state)
{
	struct msm_hardware_charger_priv *hw_chg_priv;

//	debug_print(__func__, hw_chg_removed);
	PrintLog_INFO("%s current_chg_priv=0x%x", __func__, (int)msm_chg.current_chg_priv);
	if((int)msm_chg.current_chg_priv)
		PrintLog_INFO("%s hw_chg_state=%d\n", __func__, msm_chg.current_chg_priv->hw_chg_state);

	if (msm_chg.current_chg_priv == hw_chg_removed) {
		PrintLog_DEBUG("%s msm_chg.current_chg_priv=hw_chg_removed\n", __func__);
// modified by yafi-begin
#if defined(ORG_VER)
		msm_disable_system_current(hw_chg_removed);
#else
#endif
// modified by yafi-end
		if (msm_chg.current_chg_priv->hw_chg_state
						== CHG_CHARGING_STATE) {
			PrintLog_DEBUG("%s calling msm_stop_charging()\n", __func__);
			if (msm_stop_charging(hw_chg_removed)) {
				dev_err(msm_chg.dev, "%s couldnt stop chg\n",
					msm_chg.current_chg_priv->hw_chg->name);
			}
		}
		msm_chg.current_chg_priv = NULL;
	}

	hw_chg_removed->hw_chg_state = new_state;
	PrintLog_DEBUG("%s msm_chg.current_chg_priv=0x%x\n", __func__, (int)msm_chg.current_chg_priv);
	PrintLog_DEBUG("%s msm_chg.batt_status_01=%d\n", __func__, msm_chg.batt_status);
	if (msm_chg.current_chg_priv == NULL) {
		hw_chg_priv = find_best_charger();
		if (hw_chg_priv == NULL) {
			dev_info(msm_chg.dev, "%s: no chargers\n", __func__);
			/* if the battery was Just finished charging
			 * we keep that state as is so that we dont rush
			 * in to charging the battery when a charger is
			 * plugged in shortly. */
// modified by yafi-begin
#if defined(ORG_VER)
			if (is_batt_status_charging())
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
#else
			if (is_batt_status_charging() || (msm_chg.batt_status == BATT_STATUS_JUST_FINISHED_CHARGING) || (msm_chg.batt_status == BATT_STATUS_TEMPERATURE_OUT_OF_RANGE))
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
#endif
// modified by yafi-end
		} else {
			msm_chg.current_chg_priv = hw_chg_priv;
// modified by yafi-begin
#if defined(ORG_VER)
			msm_enable_system_current(hw_chg_priv);
#else
#endif
// modified by yafi-end
			dev_info(msm_chg.dev,
				 "%s: best charger = %s\n", __func__,
				 msm_chg.current_chg_priv->hw_chg->name);

			if (!is_batt_status_capable_of_charging())
				return;

			if (msm_start_charging()) {
				/* we couldnt start charging for some reason */
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
			}
		}
	}

	PrintLog_DEBUG("%s msm_chg.batt_status_02=%d\n", __func__, msm_chg.batt_status);
	/* if we arent charging stop the safety timer */
	if (!is_batt_status_charging()) {
		dev_info(msm_chg.dev, "%s: stopping safety timer work\n",
				__func__);
		cancel_delayed_work(&msm_chg.teoc_work);
	}
}

static void handle_event(struct msm_hardware_charger *hw_chg, int event)
{
	struct msm_hardware_charger_priv *priv = NULL;

	/*
	 * if hw_chg is NULL then this event comes from non-charger
	 * parties like battery gauge
	 */

	PrintLog_DEBUG("%s hw_chg=0x%x, event=%d\n", __func__, (int)hw_chg, event);
//	PrintLog_DEBUG("%s msm_batt_gauge01=0x%x\n", __func__, (int)msm_batt_gauge);
	
	if (hw_chg)
		priv = hw_chg->charger_private;

	mutex_lock(&msm_chg.status_lock);

	switch (event) {
	case CHG_INSERTED_EVENT:
		if (priv->hw_chg_state != CHG_ABSENT_STATE) {
			dev_info(msm_chg.dev,
				 "%s insertion detected when cbl present",
				 hw_chg->name);
			break;
		}
		update_batt_status();
		if (hw_chg->type == CHG_TYPE_USB) {
			priv->hw_chg_state = CHG_PRESENT_STATE;
			notify_usb_of_the_plugin_event(priv, 1);
			if (usb_chg_current) {
				priv->max_source_current = usb_chg_current;
				usb_chg_current = 0;
				/* usb has already indicated us to charge */
				priv->hw_chg_state = CHG_READY_STATE;
				handle_charger_ready(priv);
			}
		} else {
			priv->hw_chg_state = CHG_READY_STATE;
			handle_charger_ready(priv);
		}
		break;
	case CHG_ENUMERATED_EVENT:	/* only in USB types */
		if (priv->hw_chg_state == CHG_ABSENT_STATE) {
			dev_info(msm_chg.dev, "%s enum without presence\n",
				 hw_chg->name);
			PrintLog_DEBUG("%s enum without presence",
				 hw_chg->name);
			break;
		}
		update_batt_status();
		dev_dbg(msm_chg.dev, "%s enum with %dmA to draw\n",
			 hw_chg->name, priv->max_source_current);
		PrintLog_DEBUG("%s enum with %dmA to draw",
			 hw_chg->name, priv->max_source_current);
		PrintLog_DEBUG("%s hw_chg_state=%d", __func__, priv->hw_chg_state);
		if (priv->max_source_current == 0) {
			/* usb subsystem doesnt want us to draw
			 * charging current */
			/* act as if the charge is removed */
			if (priv->hw_chg_state != CHG_PRESENT_STATE)
				handle_charger_removed(priv, CHG_PRESENT_STATE);
		} else {
			if (priv->hw_chg_state != CHG_READY_STATE) {
				priv->hw_chg_state = CHG_READY_STATE;
				handle_charger_ready(priv);
			}
		}
		break;
	case CHG_REMOVED_EVENT:
		if (priv->hw_chg_state == CHG_ABSENT_STATE) {
			dev_info(msm_chg.dev, "%s cable already removed\n",
				 hw_chg->name);
			break;
		}
		update_batt_status();
		if (hw_chg->type == CHG_TYPE_USB) {
			usb_chg_current = 0;
			notify_usb_of_the_plugin_event(priv, 0);
		}
		handle_charger_removed(priv, CHG_ABSENT_STATE);
		break;
	case CHG_DONE_EVENT:
		if (priv->hw_chg_state == CHG_CHARGING_STATE)
			handle_charging_done(priv);
		break;
	case CHG_BATT_BEGIN_FAST_CHARGING:
		/* only update if we are TRKL charging */
		if (msm_chg.batt_status == BATT_STATUS_TRKL_CHARGING)
			msm_chg.batt_status = BATT_STATUS_FAST_CHARGING;
		break;
	case CHG_BATT_NEEDS_RECHARGING:
		msm_chg.batt_status = BATT_STATUS_DISCHARGING;
		handle_battery_inserted();
		priv = msm_chg.current_chg_priv;
		break;
	case CHG_BATT_TEMP_OUTOFRANGE:
		/* the batt_temp out of range can trigger
		 * when the battery is absent */
		if (!is_battery_present()
		    && msm_chg.batt_status != BATT_STATUS_ABSENT) {
			msm_chg.batt_status = BATT_STATUS_ABSENT;
			handle_battery_removed();
			break;
		}
		if (msm_chg.batt_status == BATT_STATUS_TEMPERATURE_OUT_OF_RANGE)
			break;
		msm_chg.batt_status = BATT_STATUS_TEMPERATURE_OUT_OF_RANGE;
		handle_battery_removed();
		break;
	case CHG_BATT_TEMP_INRANGE:
		if (msm_chg.batt_status != BATT_STATUS_TEMPERATURE_OUT_OF_RANGE)
			break;
		msm_chg.batt_status = BATT_STATUS_ID_INVALID;
		/* check id */
		if (!is_battery_id_valid())
			break;
		/* assume that we are discharging from the battery
		 * and act as if the battery was inserted
		 * if a charger is present charging will be resumed */
		msm_chg.batt_status = BATT_STATUS_DISCHARGING;
		handle_battery_inserted();
		break;
	case CHG_BATT_INSERTED:
		if (msm_chg.batt_status != BATT_STATUS_ABSENT)
			break;
		/* debounce */
		if (!is_battery_present())
			break;
		msm_chg.batt_status = BATT_STATUS_ID_INVALID;
		if (!is_battery_id_valid())
			break;
		/* assume that we are discharging from the battery */
		msm_chg.batt_status = BATT_STATUS_DISCHARGING;
		/* check if a charger is present */
		handle_battery_inserted();
		break;
	case CHG_BATT_REMOVED:
		if (msm_chg.batt_status == BATT_STATUS_ABSENT)
			break;
		/* debounce */
		if (is_battery_present())
			break;
		msm_chg.batt_status = BATT_STATUS_ABSENT;
		handle_battery_removed();
		break;
	case CHG_BATT_STATUS_CHANGE:
		/* TODO  battery SOC like battery-alarm/charging-full features
		can be added here for future improvement */
		break;
	}
	dev_dbg(msm_chg.dev, "%s %d done batt_status=%d\n", __func__,
		event, msm_chg.batt_status);
	PrintLog_DEBUG("%s %d done batt_status=%d\n", __func__,
		event, msm_chg.batt_status);
//	PrintLog_DEBUG("%s msm_batt_gauge=0x%x\n", __func__, (int)msm_batt_gauge);
	/* update userspace */
	
// modified by yafi-begin
#if defined(ORG_VER)
	if (msm_batt_gauge)
#else
	if (msm_batt_gauge && (msm_psy_batt.work_wake_lock.flags & WAKE_LOCK_INITIALIZED))
#endif
// modified by yafi-end
		power_supply_changed(&msm_psy_batt);
	
// modified by yafi-begin
#if defined(ORG_VER)
	if (priv)
		power_supply_changed(&priv->psy);
#else
	if (priv)
	{
		PrintLog_DEBUG("%s max_source_current=%d\n",__func__, priv->max_source_current);
		if (priv->max_source_current == 1800)
		{
			power_supply_changed(&priv->psy_ac);
		}
		else
		{
			power_supply_changed(&priv->psy);
		}
	}
#endif
// modified by yafi-end

	mutex_unlock(&msm_chg.status_lock);
}

static int msm_chg_dequeue_event(struct msm_charger_event **event)
{
	unsigned long flags;

	spin_lock_irqsave(&msm_chg.queue_lock, flags);
	if (msm_chg.queue_count == 0) {
		spin_unlock_irqrestore(&msm_chg.queue_lock, flags);
		return -EINVAL;
	}
	*event = &msm_chg.queue[msm_chg.head];
	msm_chg.head = (msm_chg.head + 1) % MSM_CHG_MAX_EVENTS;
	pr_debug("%s dequeueing %d\n", __func__, (*event)->event);
//	PrintLog_DEBUG("%s dequeueing %d\n", __func__, (*event)->event);
	msm_chg.queue_count--;
	spin_unlock_irqrestore(&msm_chg.queue_lock, flags);
	return 0;
}

static int msm_chg_enqueue_event(struct msm_hardware_charger *hw_chg,
			enum msm_hardware_charger_event event)
{
	unsigned long flags;

	spin_lock_irqsave(&msm_chg.queue_lock, flags);
	if (msm_chg.queue_count == MSM_CHG_MAX_EVENTS) {
		spin_unlock_irqrestore(&msm_chg.queue_lock, flags);
		pr_err("%s: queue full cannot enqueue %d\n",
				__func__, event);
		return -EAGAIN;
	}
	pr_debug("%s queueing %d\n", __func__, event);
//	PrintLog_DEBUG("%s queueing %d\n", __func__, event);
	msm_chg.queue[msm_chg.tail].event = event;
	msm_chg.queue[msm_chg.tail].hw_chg = hw_chg;
	msm_chg.tail = (msm_chg.tail + 1)%MSM_CHG_MAX_EVENTS;
	msm_chg.queue_count++;
	spin_unlock_irqrestore(&msm_chg.queue_lock, flags);
	return 0;
}

static void process_events(struct work_struct *work)
{
	struct msm_charger_event *event;
	int rc;

	do {
		rc = msm_chg_dequeue_event(&event);
		if (!rc)
			handle_event(event->hw_chg, event->event);
	} while (!rc);
}

/* USB calls these to tell us how much charging current we should draw */
void msm_charger_vbus_draw(unsigned int mA)
{
	if (usb_hw_chg_priv) {
// modified by yafi-begin
#if defined(ORG_VER)
		pr_err("%s %dmA\n", __func__, mA);
#else
		PrintLog_INFO("%s %dmA\n", __func__, mA);
#endif
// modified by yafi-end
		usb_hw_chg_priv->max_source_current = mA;
		msm_charger_notify_event(usb_hw_chg_priv->hw_chg,
						CHG_ENUMERATED_EVENT);
	} else
	{
		/* remember the current, to be used when charger is ready */
		usb_chg_current = mA;
// modified by yafi-begin
#if defined(ORG_VER)
		pr_err("%s called early;charger isnt initialized %dmA\n", __func__, mA);
#else
		PrintLog_INFO("%s called early;charger isnt initialized %dmA\n", __func__, mA);
#endif
// modified by yafi-end
	}
}

static int __init determine_initial_batt_status(void)
{
	int rc;

	if (is_battery_present())
		if (is_battery_id_valid())
			if (is_battery_temp_within_range())
				msm_chg.batt_status = BATT_STATUS_DISCHARGING;
			else
				msm_chg.batt_status
				    = BATT_STATUS_TEMPERATURE_OUT_OF_RANGE;
		else
			msm_chg.batt_status = BATT_STATUS_ID_INVALID;
	else
		msm_chg.batt_status = BATT_STATUS_ABSENT;

	if (is_batt_status_capable_of_charging())
		handle_battery_inserted();

	rc = power_supply_register(msm_chg.dev, &msm_psy_batt);
	if (rc < 0) {
		dev_err(msm_chg.dev, "%s: power_supply_register failed"
			" rc=%d\n", __func__, rc);
		return rc;
	}

	/* start updaing the battery powersupply every msm_chg.update_time
	 * milliseconds */
	queue_delayed_work(msm_chg.event_wq_thread,
				&msm_chg.update_heartbeat_work,
			      round_jiffies_relative(msecs_to_jiffies
						     (msm_chg.update_time)));

	pr_debug("%s:OK batt_status=%d\n", __func__, msm_chg.batt_status);
	PrintLog_DEBUG("%s:OK batt_status=%d\n", __func__, msm_chg.batt_status);
	return 0;
}

static int __devinit msm_charger_probe(struct platform_device *pdev)
{
	msm_chg.dev = &pdev->dev;

	PrintLog_INFO("%s Begin\n", __func__);
	
	if (pdev->dev.platform_data) {

		unsigned int milli_secs;

		struct msm_charger_platform_data *pdata
		    =
		    (struct msm_charger_platform_data *)pdev->dev.platform_data;

		milli_secs = pdata->safety_time * 60 * MSEC_PER_SEC;
		if (milli_secs > jiffies_to_msecs(MAX_JIFFY_OFFSET)) {
			dev_warn(&pdev->dev, "%s: safety time too large"
				 "%dms\n", __func__, milli_secs);
			milli_secs = jiffies_to_msecs(MAX_JIFFY_OFFSET);
		}
		msm_chg.safety_time = milli_secs;

		milli_secs = pdata->update_time * 60 * MSEC_PER_SEC;
		if (milli_secs > jiffies_to_msecs(MAX_JIFFY_OFFSET)) {
			dev_warn(&pdev->dev, "%s: safety time too large"
				 "%dms\n", __func__, milli_secs);
			milli_secs = jiffies_to_msecs(MAX_JIFFY_OFFSET);
		}

// modified by yafi-begin
#if defined(ORG_VER)
		msm_chg.update_time = milli_secs;
#else
//		msm_chg.update_time = 10 * MSEC_PER_SEC;
		msm_chg.update_time = 20 * MSEC_PER_SEC;
#endif
// modified by yafi-end

		msm_chg.max_voltage = pdata->max_voltage;
		msm_chg.min_voltage = pdata->min_voltage;
		msm_chg.get_batt_capacity_percent =
		    pdata->get_batt_capacity_percent;
	}
	if (msm_chg.safety_time == 0)
		msm_chg.safety_time = CHARGING_TEOC_MS;
	if (msm_chg.update_time == 0)
		msm_chg.update_time = UPDATE_TIME_MS;
	if (msm_chg.max_voltage == 0)
		msm_chg.max_voltage = DEFAULT_BATT_MAX_V;
	if (msm_chg.min_voltage == 0)
		msm_chg.min_voltage = DEFAULT_BATT_MIN_V;
	if (msm_chg.get_batt_capacity_percent == NULL)
		msm_chg.get_batt_capacity_percent =
		    msm_chg_get_batt_capacity_percent;

	mutex_init(&msm_chg.status_lock);
	INIT_DELAYED_WORK(&msm_chg.teoc_work, teoc);
	INIT_DELAYED_WORK(&msm_chg.update_heartbeat_work, update_heartbeat);
// modified by yafi-begin
#if defined(ORG_VER)
#else
	INIT_DELAYED_WORK(&msm_chg.update_battery_status_work, update_battery_status);
#endif
// modified by yafi-end	

	wake_lock_init(&msm_chg.wl, WAKE_LOCK_SUSPEND, "msm_charger");
	PrintLog_INFO("%s End\n", __func__);

	return 0;
}

static int __devexit msm_charger_remove(struct platform_device *pdev)
{
	wake_lock_destroy(&msm_chg.wl);
	mutex_destroy(&msm_chg.status_lock);
	power_supply_unregister(&msm_psy_batt);
	return 0;
}

// modified by yafi-begin
#if defined(ORG_VER)
#else
extern bool smb137b_disable_charger_timeout(bool disable);
bool msm_charger_timeout_disable(bool disable)
{
	if (disable)
	{
		gmsm_charger_timeout_disable = true;
		cancel_delayed_work(&msm_chg.teoc_work);
		if (smb137b_disable_charger_timeout(true))
			return true;
	}
	else
	{
		gmsm_charger_timeout_disable = false;
		if (is_batt_status_charging())
			queue_delayed_work(msm_chg.event_wq_thread,
								&msm_chg.teoc_work,
								round_jiffies_relative(msecs_to_jiffies(msm_chg.safety_time)));
		if (smb137b_disable_charger_timeout(false))
			return true;
	}
	return false;
}

EXPORT_SYMBOL(msm_charger_timeout_disable);
#endif
// modified by yafi-end

int msm_charger_notify_event(struct msm_hardware_charger *hw_chg,
			     enum msm_hardware_charger_event event)
{
	msm_chg_enqueue_event(hw_chg, event);
	queue_work(msm_chg.event_wq_thread, &msm_chg.queue_work);
	return 0;
}
EXPORT_SYMBOL(msm_charger_notify_event);

int msm_charger_register(struct msm_hardware_charger *hw_chg)
{
	struct msm_hardware_charger_priv *priv;
	int rc = 0;

	if (!msm_chg.inited) {
		pr_err("%s: msm_chg is NULL,Too early to register\n", __func__);
		return -EAGAIN;
	}

	if (hw_chg->start_charging == NULL
		|| hw_chg->stop_charging == NULL
		|| hw_chg->name == NULL
		|| hw_chg->rating == 0) {
		pr_err("%s: invalid hw_chg\n", __func__);
		return -EINVAL;
	}

	priv = kzalloc(sizeof *priv, GFP_KERNEL);
	if (priv == NULL) {
		dev_err(msm_chg.dev, "%s kzalloc failed\n", __func__);
		return -ENOMEM;
	}

// modified by yafi-begin
#if defined(ORG_VER)
#else
	priv->psy_ac.name = "ac";
	priv->psy_ac.type = POWER_SUPPLY_TYPE_MAINS;

	priv->psy_ac.supplied_to = msm_power_supplied_to;
	priv->psy_ac.num_supplicants = ARRAY_SIZE(msm_power_supplied_to);
	priv->psy_ac.properties = msm_power_props;
	priv->psy_ac.num_properties = ARRAY_SIZE(msm_power_props);
	priv->psy_ac.get_property = msm_power_get_property_ac;

	rc = power_supply_register(NULL, &priv->psy_ac);
	if (rc) {
		dev_err(msm_chg.dev, "%s power_supply_register AC failed\n",
			__func__);
		goto out;
	}
#endif
// modified by yafi-end

// modified by yafi-begin
#if defined(ORG_VER)
	priv->psy.name = hw_chg->name;
	if (hw_chg->type == CHG_TYPE_USB)
		priv->psy.type = POWER_SUPPLY_TYPE_USB;
	else
		priv->psy.type = POWER_SUPPLY_TYPE_MAINS;
#else
	if (hw_chg->type == CHG_TYPE_USB)
	{
		priv->psy.name = "usb";
		priv->psy.type = POWER_SUPPLY_TYPE_USB;
	}
	else
	{
		priv->psy.name = "ac";
		priv->psy.type = POWER_SUPPLY_TYPE_MAINS;
	}
#endif
// modified by yafi-end

	priv->psy.supplied_to = msm_power_supplied_to;
	priv->psy.num_supplicants = ARRAY_SIZE(msm_power_supplied_to);
	priv->psy.properties = msm_power_props;
	priv->psy.num_properties = ARRAY_SIZE(msm_power_props);
	priv->psy.get_property = msm_power_get_property;

	rc = power_supply_register(NULL, &priv->psy);
	if (rc) {
		dev_err(msm_chg.dev, "%s power_supply_register USB failed\n",
			__func__);
		goto out;
	}

	priv->hw_chg = hw_chg;
	priv->hw_chg_state = CHG_ABSENT_STATE;
	INIT_LIST_HEAD(&priv->list);
	mutex_lock(&msm_chg.msm_hardware_chargers_lock);
	list_add_tail(&priv->list, &msm_chg.msm_hardware_chargers);
	mutex_unlock(&msm_chg.msm_hardware_chargers_lock);
	hw_chg->charger_private = (void *)priv;
	return 0;

out:
// modified by yafi-begin
#if defined(ORG_VER)
#else
	wake_lock_destroy(&msm_chg.wl);
#endif
// modified by yafi-end
	kfree(priv);
	return rc;
}
EXPORT_SYMBOL(msm_charger_register);

void msm_battery_gauge_register(struct msm_battery_gauge *batt_gauge)
{
	PrintLog_DEBUG("%s entering bg=0x%x\n", __func__, (int)batt_gauge);
	
	if (msm_batt_gauge) {
		msm_batt_gauge = batt_gauge;
		pr_err("msm-charger %s multiple battery gauge called\n",
								__func__);
	} else {
		msm_batt_gauge = batt_gauge;
		determine_initial_batt_status();
	}
}
EXPORT_SYMBOL(msm_battery_gauge_register);

void msm_battery_gauge_unregister(struct msm_battery_gauge *batt_gauge)
{
	msm_batt_gauge = NULL;
}
EXPORT_SYMBOL(msm_battery_gauge_unregister);

int msm_charger_unregister(struct msm_hardware_charger *hw_chg)
{
	struct msm_hardware_charger_priv *priv;

	priv = (struct msm_hardware_charger_priv *)(hw_chg->charger_private);
	mutex_lock(&msm_chg.msm_hardware_chargers_lock);
	list_del(&priv->list);
	mutex_unlock(&msm_chg.msm_hardware_chargers_lock);
// modified by yafi-begin
#if defined(ORG_VER)
#else
	wake_lock_destroy(&msm_chg.wl);
#endif
// modified by yafi-end
	power_supply_unregister(&priv->psy);
	kfree(priv);
	return 0;
}
EXPORT_SYMBOL(msm_charger_unregister);

static int msm_charger_suspend(struct device *dev)
{
	dev_dbg(msm_chg.dev, "%s suspended\n", __func__);
	PrintLog_INFO("%s suspended\n", __func__);
	msm_chg.stop_update = 1;
	cancel_delayed_work(&msm_chg.update_heartbeat_work);
	mutex_lock(&msm_chg.status_lock);
	handle_battery_removed();
	mutex_unlock(&msm_chg.status_lock);
	return 0;
}

static int msm_charger_resume(struct device *dev)
{
	dev_dbg(msm_chg.dev, "%s resumed\n", __func__);
	PrintLog_INFO("%s resumed\n", __func__);
	msm_chg.stop_update = 0;
	/* start updaing the battery powersupply every msm_chg.update_time
	 * milliseconds */
	queue_delayed_work(msm_chg.event_wq_thread,
				&msm_chg.update_heartbeat_work,
			      round_jiffies_relative(msecs_to_jiffies
						     (msm_chg.update_time)));
	mutex_lock(&msm_chg.status_lock);
	handle_battery_inserted();
	mutex_unlock(&msm_chg.status_lock);
	return 0;
}

static SIMPLE_DEV_PM_OPS(msm_charger_pm_ops,
		msm_charger_suspend, msm_charger_resume);

static struct platform_driver msm_charger_driver = {
	.probe = msm_charger_probe,
	.remove = __devexit_p(msm_charger_remove),
	.driver = {
		   .name = "msm-charger",
		   .owner = THIS_MODULE,
		   .pm = &msm_charger_pm_ops,
	},
};

static int __init msm_charger_init(void)
{
	int rc;

	PrintLog_INFO("%s Begin\n", __func__);

	INIT_LIST_HEAD(&msm_chg.msm_hardware_chargers);
	msm_chg.count_chargers = 0;
	mutex_init(&msm_chg.msm_hardware_chargers_lock);

	msm_chg.queue = kzalloc(sizeof(struct msm_charger_event)
				* MSM_CHG_MAX_EVENTS,
				GFP_KERNEL);
	if (!msm_chg.queue) {
		rc = -ENOMEM;
		goto out;
	}
	msm_chg.tail = 0;
	msm_chg.head = 0;
	spin_lock_init(&msm_chg.queue_lock);
	msm_chg.queue_count = 0;
	INIT_WORK(&msm_chg.queue_work, process_events);
	msm_chg.event_wq_thread = create_workqueue("msm_charger_eventd");
	if (!msm_chg.event_wq_thread) {
		rc = -ENOMEM;
		goto free_queue;
	}
	rc = platform_driver_register(&msm_charger_driver);
	if (rc < 0) {
		pr_err("%s: FAIL: platform_driver_register. rc = %d\n",
		       __func__, rc);
		goto destroy_wq_thread;
	}
	msm_chg.inited = 1;
	PrintLog_INFO("%s End\n", __func__);
	return 0;

destroy_wq_thread:
	destroy_workqueue(msm_chg.event_wq_thread);
free_queue:
	kfree(msm_chg.queue);
out:
	return rc;
}

static void __exit msm_charger_exit(void)
{
	flush_workqueue(msm_chg.event_wq_thread);
	destroy_workqueue(msm_chg.event_wq_thread);
	kfree(msm_chg.queue);
	platform_driver_unregister(&msm_charger_driver);
}

module_init(msm_charger_init);
module_exit(msm_charger_exit);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Abhijeet Dharmapurikar <adharmap@codeaurora.org>");
MODULE_DESCRIPTION("Battery driver for Qualcomm MSM chipsets.");
MODULE_VERSION("1.0");
