/*
 *  Convert a logo in ASCII RLE format to C source suitable for inclusion in
 *  the Linux kernel
 *
 */

#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


static const char *programname;
static const char *filename;
static const char *logoname = "init_logo";
static const char *outputname;
static FILE *out;

static void die(const char *fmt, ...)
    __attribute__ ((noreturn)) __attribute ((format (printf, 1, 2)));
static void usage(void) __attribute ((noreturn));

static unsigned char *logo_data;
static unsigned int maxval = 0;
static int write_hex_cnt;

static void read_image(void)
{
    FILE *fp;
    unsigned int i;
	
    /* open image file */
    fp = fopen(filename, "rb");
    if (!fp){
	    die("Cannot open file %s: %s\n", filename, strerror(errno));
    }
    /* seek to end of file */
    fseek(fp, (off_t)0, 2);

    /* get the length of file */
    maxval = (unsigned int)ftell(fp);
	
    /* allocate image data */
    logo_data = (unsigned char*)malloc(maxval);
    if (!logo_data){
	    die("%s\n", strerror(errno));
    }
    /* seek to start of file */
    fseek(fp, (off_t)0, 0);

    /* get image data */
    for (i = 0; i < maxval; i++) {
	    logo_data[i] =  fgetc(fp);
    }

    /* close file */
    fclose(fp);

    return ;
}


static void write_header(void)
{
    /* open logo file */
    if (outputname) {
	    out = fopen(outputname, "w");
	    if (!out){
	        die("Cannot create file %s: %s\n", outputname, strerror(errno));
        }
    } else {
	    out = stdout;
    }

    fputs("/*\n", out);
    fputs(" *  DO NOT EDIT THIS FILE!\n", out);
    fputs(" *\n", out);
    fprintf(out, " *  It was automatically generated from %s\n", filename);
    fputs(" *\n", out);
    fprintf(out, " *  logo :%s\n", logoname);
    fputs(" */\n\n", out);
    //fputs("#include <linux/linux_logo.h>\n\n", out);
    fprintf(out, " unsigned int %s_data[]  = {\n",
	    logoname);
}

static void write_footer(void)
{
    fputs("\n};\n\n", out);
	
    fprintf(out, " unsigned int init_logo_size  = %d ; \n\n",
	    maxval);

    /* close logo file */
    if (outputname){
	    fclose(out);
    }

    if(logo_data){
	    free(logo_data);
    }
}



static void write_hex2( unsigned int val)
{
    if (write_hex_cnt % 8){
	    fprintf(out, ", 0x%08x", val);
    }
    else if (write_hex_cnt){
	    fprintf(out, ",\n\t0x%08x", val);
    }
    else{
	    fprintf(out, "\t0x%08x", val);
    }
    write_hex_cnt++;
}

static void write_logo_mono(void)
{
    unsigned int i = 0;

    /* write file header */
    write_header();

    /* write logo data */
    while (i < maxval) {
	    write_hex2((unsigned int)((logo_data[i]<<24)|(logo_data[i+1]<<16)| (logo_data[i+2]<<8)) | (logo_data[i+3]));		
	    i+=4;
    }

    /* write logo structure and file footer */
    write_footer();
}


static void die(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);

    exit(1);
}

static void usage(void)
{
    die("\n"
	"Usage: %s [options] <filename>\n"
	"\n"
	"Valid options:\n"
	"    -h          : display this usage information\n"
	"    -n <name>   : specify logo name (default: init_logo)\n"
	"    -o <output> : output to file <output> instead of stdout\n"
	"\n", programname);
}

int main(int argc, char *argv[])
{
    int opt;

    programname = argv[0];

    opterr = 0;
    while (1){
		opt = getopt(argc, argv, "hn:o:t:");
		if (opt == -1)
	   		 break;

		switch (opt) {
	    	case 'h':
			usage();
			break;

	    	case 'n':
			logoname = optarg;
			break;

	    	case 'o':
			outputname = optarg;
			break;

	    	default:
			usage();
			break;
		}
    }
    if (optind != argc-1)
   	{
		usage();
	}

    filename = argv[optind];

    read_image();

    write_logo_mono();
	   
    exit(0);
}

